#!/usr/bin/python
#-----------------------------------------------------------------------
# File    : special.py
# Contents: accretion/frequent item set mining: special cases
# Author  : Christian Borgelt
# History : 2012.02.06 file created as accfim.py
#           2012.02.07 tests for false negatives completed
#           2012.05.30 adapted to different set of experiments
#           2012.07.02 command list setup made separate function
#           2013.01.16 adapted to psts.py and fim library
#           2014.05.27 modules placed in sibling directories
#-----------------------------------------------------------------------
from sys     import argv, path
from os      import makedirs
from os.path import join
from time    import strftime
from random  import seed
UTILDIR  = join('..', 'util')
if UTILDIR  not in path: path.insert(0, UTILDIR)
from id2par  import get_args, project, run
NEURODIR = join('..', 'neuro')
if NEURODIR not in path: path.insert(0, NEURODIR)
from psts    import genpsts, psts2tras
from fim     import apriori

#-----------------------------------------------------------------------

def get_cmds (target='c'):      # get dictionary of python functions
    stats = ['p']               # list of statistics
    coma  = [('target', target), ('supp',-2), ('zmin', 2),
             ('agg', 'n'), ('thresh', 1), ('mode', 'z'),
             ('report', 'a')]   # get the common arguments
    fim   = [('a+3', apriori, [('prune', +3)] +coma)]
    return dict([(n+'_'+s, (c, dict(a +[('eval', s)])))
                 for s in stats for n,c,a in fim])

#-----------------------------------------------------------------------

def fneg (ncnt=100, period=3, rate=20, width=0.003, size=5, coin=5,
          target='c', runs=1000, verbose=100, srng=1):
    srng += 1023*size           # init. seed and get data file name
    name  = 'n%d_p%g_r%g_w%g_%c_%d_%d' \
          % (ncnt,period,rate,width*1000,target,size,coin)
    cmds  = get_cmds(target)    # get the different mining commands
    dect  = dict([(c, 0.0) for c in cmds])
    excs  = dict([(c, 0.0) for c in cmds])
    asm   = set(range(1,size+1))# create assembly to inject
    for i in range(runs):       # repeat the experiments multiple times
        if verbose > 0 and i % verbose == 0:
            print(strftime('%Y.%m.%d %H:%M:%S special ')
                  +name +' ' +str(i))
        seed(srng)              # seed random number generator
        psts = genpsts(ncnt, period, rate, asms=[[asm,-coin]])
        tras = psts2tras(psts, width, period)
        for c in cmds:          # generate trains and traverse commands
            fn,args = cmds[c]   # mine the generated spike trains
            pats = fn(tras, **args)
            d = e = 0           # initialize the detection flags
            for p,r in pats:    # and traverse the found patterns
                if len(p) >= size and set(p) >= asm:
                    if len(p) > size: e = 1
                    d = 1       # set detection and excess neuron flags
            dect[c] += d        # count the correct detections and
            excs[c] += e        # detections with excess neurons
        srng += coin            # switch to the next random seed
    if verbose > 0:             # print termination log message
        print(strftime('%Y.%m.%d %H:%M:%S special ')
              +name +' ' +str(runs) +' done.')
    try: makedirs('pats')       # ensure that a directory exists
    except OSError: pass        # for the pattern (spectrum) files
    n = 1.0/runs if runs > 0 else 1
    for c in cmds:              # traverse the different commands
        with open(join('data','fn_'+c+'_'+name+'.txt'), 'w') as out:
            d = dect[c]         # write pattern counters to file
            out.write('%f '  % (1.0-d*n))
            out.write('%f\n' % (excs[c]/float(d) if d > 0 else 0.0))
    return (dect,excs)          # return the pattern counters

#-----------------------------------------------------------------------

def collect (ncnt=100, period=3, rate=20, width=0.003,
             sizes=5, coins=5, target='c'):
    name = 'n%d_p%g_r%g_w%g_%c' % (ncnt,period,rate,width*1000,target)
    dect = dict()           # get the pattern count file name base
    excs = dict()           # and initialize the dictionaries
    cmds = get_cmds(target) # get the different mining commands
    for c in cmds:          # traverse the commands
        for s in [(z,n) for z in sizes for n in coins]:
            inpfn = join('data','fn_'+c+'_'+name+('_%d_%d.txt' %s))
            with open(inpfn, 'r') as inp:
                dect[s],excs[s] = inp.read()[:-1].split()
        with open(join('data','fn_'+c+'_'+name+'.txt'), 'w') as out:
            for x in sorted([(k[0], k[1], dect[k]) for k in dect]):
                out.write('%g %g %s\n' % x)
        with open(join('data','xs_'+c+'_'+name+'.txt'), 'w') as out:
            for x in sorted([(k[0], k[1], excs[k]) for k in excs]):
                out.write('%g %g %s\n' % x)

#-----------------------------------------------------------------------

if __name__ == '__main__':
    grids = [ [ ( 'ncnt',   [  100 ] ),
                ( 'period', [    3 ] ),
                ( 'rate',   [ 30, 40 ] ),
                ( 'width',  [ 0.005, 0.004, 0.003 ] ),
                ( 'size',   [ 2, 3, 4, 5, 6, 7, 8, 9 ] ),
                ( 'coin',   [ 2, 3, 4, 5, 6, 7, 8, 9 ] ),
                ( 'target', [  'm' ] ),
                ( 'runs',   [ 1000 ] ) ] ]

    if len(argv) < 2 or argv[1] != 'collect':
        run(fneg, grids, int(argv[1])-1 if len(argv) > 1 else -1)
    else:
        ranges = dict(project(grids, ['size','coin'])[0])
        grids  = project(grids, ['ncnt','period','rate','width',
                                 'target','rfnid'])
        for i in range(get_args(grids)):
            collect(sizes=ranges['size'], coins=ranges['coin'],
                    **get_args(grids, i))
